(*	importing Individuals in Tab-Delimited File script
	GEDitCOM II Apple Script
	7 JAN 2009, by John A. Nairn

	This script will import any tab-delimited file with data for individuals
	into the current front document. See details in help for this extension
*)

-- key properties and variables
property scriptName : "Import Tab Delimited Script"
property fieldDelim : tab
property lineDelims : {ASCII character 10, (ASCII character 10) & (ASCII character 13)}
global impData, columns, recLine, tags, addChil, numFams

-- get file name
copy (choose file with prompt "Select Tab-Delimited File to Import") to txtfile
open for access txtfile
set impData to read txtfile using delimiter lineDelims
close access txtfile

-- must have at least the header line and 1 record line
set numLines to number of items in impData
if numLines < 2 then
	-- but what if used CR line feeds
	set lineDelims to {ASCII character 13}
	open for access txtfile
	set impData to read txtfile using delimiter lineDelims
	close access txtfile
	set numLines to number of items in impData
	if numLines < 2 then
		tell application "GEDitCOM II"
			user option title "No data could be found in that file." buttons {"OK"}
		end tell
		return
	end if
end if
-- faster to use a reference to the read a list according to Apple
set refData to a reference to impData

-- first line is list of tags
-- get first line and verify delimiter
set tagLine to item 1 of refData
set toff to offset of tab in tagLine
if toff is 0 then
	set fieldDelim to ","
end if

-- decode into list of lists of tags with one for each column
-- for example { {NAME}, {BIRT, DATE}, {BIRT, PLAC} } in final tags
set tags to {}
set the text item delimiters to {fieldDelim}
set columns to number of text items in tagLine
repeat with i from 1 to columns
	set tagExpr to text item i of tagLine
	if tagExpr is "" then
		exit repeat
	end if
	set the text item delimiters to {"."}
	set numTags to number of text items in tagExpr
	set tagList to {}
	repeat with j from 1 to numTags
		set end of tagList to text item j of tagExpr
	end repeat
	set the text item delimiters to {fieldDelim}
	set end of tags to tagList
end repeat
set columns to number of items in tags

-- list of individual created before their line is reached
set addChil to {}
set numFams to 0
tell application "GEDitCOM II"
	make new document
end tell

-- decode each line read from the file
repeat with i from 2 to numLines
	set recLine to item i of refData
	tell application "GEDitCOM II"
		tell front document
			set newID to "@I" & i & "@"
			set newIndi to make new individual with properties {id:newID}
			
			-- fill the individual, quit on error
			if my fillRecord(newIndi, i) is false then
				return
			end if
		end tell
	end tell
end repeat

-- add CHIL links and reparse GEDCOM if has places
tell application "GEDitCOM II"
	
	-- add CHIL links
	set nchil to number of items in addChil
	if nchil > 0 then
		tell front document
			repeat with i from 1 to nchil by 2
				set chilID to item i of addChil
				set parID to item (i + 1) of addChil
				
				-- get FAMS link from parent
				set famid to ""
				try
					set famsref to (a reference to structure named "FAMS" of individual id parID)
					set famid to contents of famsref
				end try
				
				-- set links if found
				if famid is not "" then
					tell family id famid
						make new structure with properties {name:"CHIL", contents:chilID}
					end tell
					tell individual id chilID
						make new structure with properties {name:"FAMC", contents:famid}
					end tell
				end if
			end repeat
		end tell
	end if
	
	-- reparce records with PLAC tags
	tell front document
		my ReparseRecords(every individual)
		my ReparseRecords(every family)
	end tell
	
	-- count (not header and submitter)
	set numLines to (count of gedcomRecords in front document) - 2
end tell

-- alert when done
if numLines is 1 then
	set msg to "1 individual"
	set msg2 to "This record"
else
	set msg to ((numLines as string) & " individuals, families and places") as string
	set msg2 to "These records"
end if
tell application "GEDitCOM II"
	set msg2 to msg2 & " can now be merged or dragged to any other file."
	user option title "Custom importing of " & msg & " was completed." message msg2 buttons {"OK"}
end tell
return

(* This subrotine will fill the new record referenced by recRef with the
	data in the global variable recLine. It will use the global variable
	tags to determine where to put the data
*)
on fillRecord(recRef, lineNum)
	-- skip empty lines
	if recLine is "" then
		DeleteIndi(recRef)
		return true
	end if
	
	-- parce columns of data removing quotes and other tasks
	set lineData to every text item of recLine
	set lineList to {}
	set text item delimiters to {""}
	set ncol to number of items in lineData
	set hasData to false
	set inQuote to false
	repeat with ii from 1 to ncol
		set colData to item ii of lineData
		set nc to number of characters in colData
		if nc > 0 then
			if fieldDelim is tab then
				-- tab delim - strip quotes if found
				if first character of colData is quote and last character of colData is quote then
					get characters 2 thru (nc - 1) of colData
					set colData to the result as string
				end if
			else
				-- CSV - need to reconstruct those with commas between quotes
				if inQuote is false then
					if first character of colData is quote then
						if nc > 1 then
							if last character of colData is quote then
								get characters 2 thru (nc - 1) of colData as string
								set colData to the result as string
							else
								get characters 2 thru nc of colData as string
								set qData to the result as string
								set inQuote to true
							end if
						else
							set qData to ""
							set inQuote to true
						end if
					end if
				else
					if last character of colData is quote then
						if nc > 1 then
							get characters 1 thru (nc - 1) of colData as string
							set colData to the result as string
						else
							set colData to ""
						end if
						set colData to qData & "," & colData
						set inQuote to false
					else
						set qData to qData & "," & colData
					end if
				end if
			end if
		end if
		if inQuote is false then
			set end of lineList to colData
			if colData is not "" then set hasData to true
			if number of items in lineList  columns then exit repeat
		end if
	end repeat
	if inQuote is true then
		set end of lineList to qData
		if qData is not "" then set hasData to true
	end if
	set text item delimiters to {fieldDelim}
	
	-- skip lines with not data in any column
	if hasData is false then
		DeleteIndi(recRef)
		return true
	end if
	
	-- blank unless has FAMS data
	set famsID to ""
	
	tell application "GEDitCOM II"
		repeat with jj from 1 to columns
			-- read the next column of data
			try
				set recData to item jj of lineList
			on error
				set recData to ""
			end try
			
			-- insert data if it is not empty
			if recData is not "" then
				-- get the tags for this column
				set tagList to item jj of tags
				set numTags to number of items in tagList
				set tag1 to item 1 of tagList
				
				-- special case for single NAME tag
				if tag1 is "NAME" and numTags is 1 then
					tell recRef
						set name to recData
					end tell
					
					-- special case for FAMC tag
				else if tag1 is "FAMC" then
					set end of addChil to id of recRef
					set end of addChil to "@I" & recData & "@"
					
				else
					-- get structure to set in current individual
					set sref to my GetStructure(tag1, recRef)
					set useRecRef to recRef
					set nextTag to 2
					
					-- need to trap those referencing a linked family records
					
					-- FAMS.spse means husband or wife of the individual, value should be their line number
					-- FAMS.tags set tags in the family record
					-- Only one FAMS per individual (i.e., cannot do multiple spouses)
					if tag1 is "FAMS" then
						-- error if only one
						if numTags = 1 then
							set msg to "Column headers with FAMS only are not allowed"
							user option title msg buttons {"OK"}
							return false
						end if
						
						-- save this individual details
						set indiid to id of recRef
						set indiSex to sex of recRef
						
						-- get second tag
						set tag2 to item 2 of tagList -- maybe spse or other
						
						-- if don't have family link, look in the spouse record
						-- but only work sif it was created already
						-- Create one if needed
						-- Note: use 0 or 1 if no spouse but create family anyway
						if (famsID is "") and (tag2 is "spse") then
							set spseID to "@I" & recData & "@"
							tell front document
								try
									set famsref to (a reference to structure named "FAMS" of individual id spseID)
									set famsID to contents of famsref
								end try
							end tell
						end if
						
						-- if family was not created before, create it now if a spouse tag
						-- In other words, FAMS.spse must be first FAMS column
						if (famsID is "") and (tag2 is "spse") then
							tell front document
								set linkfam to make new family
								set numFams to numFams + 1
								set famsID to id of linkfam
							end tell
							
							-- set contents of link in recRef (current individual) to point to new family
							set contents of sref to famsID
							
							-- writing to the family record now
							set useRecRef to linkfam
						else if famsID is not "" then
							if tag2 is "spse" then
								-- second spouse will find known famsID
								-- set contents of link in recRef (current individual) to point to new family
								set contents of sref to famsID
							end if
							set useRecRef to family id famsID of front document
						else
							-- error found FAMS.other before family was created
							set msg to "The first FAMS column with data for an individual must be FAMS.spse"
							user option title "A family record as spouse was referenced before the family was created" message msg buttons {"OK"}
							return false
						end if
						
						-- if needed, set the backtag
						if tag2 is "spse" then
							if indiSex is "M" then
								set backTag to "HUSB"
								set altTag to "WIFE"
							else
								set backTag to "WIFE"
								set altTag to "HUSB"
							end if
							tell useRecRef
								-- if already has backTag, swith to altTag
								-- this check allows same-sex couple
								set currentLink to find structures tag backTag
								if number of items in currentLink > 0 then
									set backTag to altTag
								end if
								set linkref to (make new structure with properties {name:backTag})
								set contents of linkref to indiid
							end tell
							-- tags after spse are ignored
							set useRecRef to ""
						else
							-- get the structure in the family record
							set sref to my GetStructure(tag2, useRecRef)
							set nextTag to 3
						end if
					end if
					
					-- set contents of sref (if not set already) or find subordinate structure
					-- to set if there are remaining tags in tagList
					if useRecRef is not "" then
						repeat with k from nextTag to numTags
							set tagk to item k of tagList
							tell useRecRef
								try
									set subref to (a reference to structure named tagk of sref)
									name of subref
								on error
									tell sref
										set subref to (make new structure with properties {name:tagk})
									end tell
								end try
							end tell
							set sref to subref
						end repeat
						
						-- set the contents of the new structure
						set contents of sref to recData
					end if
					
				end if -- end of tag options
			end if -- end of recDate not ""
			
		end repeat
	end tell
	
	return true
end fillRecord

-- delete individual if not needed
on DeleteIndi(recRef)
	tell application "GEDitCOM II"
		tell front document
			delete recRef
		end tell
	end tell
end DeleteIndi

-- get reference to structure with given tag in recRef
-- create the structure if not ther
on GetStructure(tag1, recRef)
	tell application "GEDitCOM II"
		tell recRef
			try
				set sref to (a reference to structure named tag1)
				name of sref
			on error
				set sref to (make new structure with properties {name:tag1})
			end try
		end tell
	end tell
	return sref
end GetStructure

-- when done reparse records that have places
on ReparseRecords(theRecs)
	tell application "GEDitCOM II"
		set nr to number of items in theRecs
		repeat with i from 1 to nr
			tell item i of the theRecs
				set ptags to find structures tag "PLAC"
				if number of items in ptags > 0 then
					set gc to gedcom
					set gedcom to gc
				end if
			end tell
		end repeat
	end tell
end ReparseRecords


